use std::fmt::{Display, Formatter};
use std::str::FromStr;

use piet::kurbo::{Rect, Vec2};
use serde_with::{DeserializeFromStr, SerializeDisplay};
use thiserror::Error;

#[derive(Debug, SerializeDisplay, DeserializeFromStr, Eq, PartialEq, Hash)]
pub struct UIntVec2 {
    x: u64,
    y: u64,
}

impl UIntVec2 {
    fn to_kurbo_vec2(&self) -> Vec2 {
        Vec2::new(self.x as f64, self.y as f64)
    }
}

#[derive(Debug, Error)]
#[error("The input value does not match the required format of <x> and <y> separated by an 'x'")]
pub struct UIntVec2FromStrError {}

impl FromStr for UIntVec2 {
    type Err = UIntVec2FromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let values = s.split_once('x');

        if let Some((x, y)) = values {
            if let Ok(x) = u64::from_str(x) {
                if let Ok(y) = u64::from_str(y) {
                    return Ok(UIntVec2 { x, y });
                }
            }
        }

        Err(UIntVec2FromStrError {})
    }
}

impl Display for UIntVec2 {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}x{}", self.x, self.y))
    }
}

pub fn parse_positive_rect_from_str(s: &str) -> Result<Rect, ()> {
    let (pairs, is_corner_points_mode) = if let Some(pairs) = s.split_once('+') {
        (pairs, false)
    } else if let Some(pairs) = s.split_once('-') {
        (pairs, true)
    } else {
        return Err(());
    };

    let first_vec = UIntVec2::from_str(pairs.0).map_err(|_| ())?.to_kurbo_vec2();
    let second_vec = UIntVec2::from_str(pairs.1).map_err(|_| ())?.to_kurbo_vec2();

    Ok(if is_corner_points_mode {
        Rect::from_points(first_vec.to_point(), second_vec.to_point())
    } else {
        Rect::from_origin_size(first_vec.to_point(), second_vec.to_size())
    })
}
