use std::path::PathBuf;
use std::str::FromStr;

use serde_with::DeserializeFromStr;
use thiserror::Error;

use crate::model::image_filter::{ImageFilter, ImageFilterFromStringError};

#[derive(Debug, Default, Clone, DeserializeFromStr)]
pub struct IconDescriptor {
    pub source: IconDescriptorSource,
    pub filter: Option<ImageFilter>,
}

#[derive(Debug, Error)]
pub enum IconDescriptorFromStrError {
    #[error("Not a valid icon identifier: {0}")]
    InvalidIconPackSource(String),

    #[error("The image filter is invalid: {0}")]
    InvalidImageFilter(#[source] ImageFilterFromStringError),

    #[error("The image filter is missing the closing ']'")]
    MissingImageFilterClosingBracket,
}

impl FromStr for IconDescriptor {
    type Err = IconDescriptorFromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let (raw_source, raw_filter) = s.split_once('[').unwrap_or((s, ""));

        let source = if let Some(raw_source) = raw_source.strip_prefix('@') {
            let (pack_id, icon_id) = raw_source
                .split_once('/')
                .ok_or_else(|| IconDescriptorFromStrError::InvalidIconPackSource(raw_source.to_string()))?;
            IconDescriptorSource::IconPack {
                pack_id: pack_id.to_owned(),
                icon_id: icon_id.to_owned(),
            }
        } else {
            IconDescriptorSource::Path(PathBuf::from(raw_source))
        };

        let filter: Option<ImageFilter> = if raw_filter.is_empty() {
            None
        } else {
            let mut raw_filter = raw_filter.to_owned();
            if raw_filter.pop().expect("emptiness was eliminated a few lines earlier") != ']' {
                return Err(IconDescriptorFromStrError::MissingImageFilterClosingBracket);
            }

            Some(ImageFilter::from_str(&raw_filter).map_err(IconDescriptorFromStrError::InvalidImageFilter)?)
        };

        Ok(IconDescriptor { source, filter })
    }
}

#[derive(Debug, Default, Clone)]
pub enum IconDescriptorSource {
    #[default]
    None,
    IconPack {
        pack_id: String,
        icon_id: String,
    },
    Path(PathBuf),
}
