use std::str::FromStr;

use piet::kurbo::Rect;
use rgb::RGB8;
use serde_with::DeserializeFromStr;
use thiserror::Error;

use crate::model::geometry::parse_positive_rect_from_str;
use crate::model::rgb::parse_rgb8_from_hex_str;

#[derive(Debug, Clone, DeserializeFromStr)]
pub struct ImageFilter {
    pub crop_original: Option<Rect>, // applied before scale and rotate
    pub scale: f32,
    pub clockwise_quarter_rotations: u8,
    pub crop: Option<Rect>, // applied after scale and rotate
    pub color: Option<RGB8>,
    pub alpha: f32,
    pub blur: f32,
    pub grayscale: bool,
    pub invert: bool,
}

#[derive(Debug, Error)]
pub enum ImageFilterFromStringError {
    #[error("Unknown filter: {name}")]
    UnknownFilter { name: String },

    #[error("Filter {name} can only be used once.")]
    FilterUsedMoreThanOnce { name: String },

    #[error("Only the following values are for rotate: -90, 90, 180, 270")]
    RotationNotAllowed,

    #[error("Filter {filter_name} requires a value ({filter_name}=<value>).")]
    FilterValueMissing { filter_name: String },

    #[error("Filter {filter_name} does not require or allow a value.")]
    FilterValueIgnored { filter_name: String },

    #[error("The following value supplied to {filter_name} could not be parsed: {raw_value}")]
    FilterValueNotParsable { filter_name: String, raw_value: String },

    #[error("{value} is not in the range ({range}) required by {filter_name}.")]
    FilterValueNotInRange { filter_name: String, value: String, range: String },
}

impl FromStr for ImageFilter {
    type Err = ImageFilterFromStringError;

    fn from_str<'a>(s: &str) -> Result<Self, Self::Err> {
        fn parse_rect_filter_value(filter_name: &str, raw_value: String) -> Result<Rect, ImageFilterFromStringError> {
            parse_positive_rect_from_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                filter_name: filter_name.to_string(),
                raw_value,
            })
        }

        fn parse_f32_filter_value(
            filter_name: &str,
            raw_value: String,
            range_string: &str,
            check_range: Box<dyn Fn(&f32) -> bool>,
        ) -> Result<f32, ImageFilterFromStringError> {
            let value = f32::from_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                filter_name: filter_name.to_string(),
                raw_value,
            })?;

            if !check_range(&value) {
                return Err(ImageFilterFromStringError::FilterValueNotInRange {
                    filter_name: filter_name.to_string(),
                    range: range_string.to_string(),
                    value: value.to_string(),
                });
            }

            Ok(value)
        }

        let filters: Vec<&str> = s.split('|').map(|f| f.trim()).collect();

        let mut result = ImageFilter {
            crop_original: None,
            scale: 1.0,
            clockwise_quarter_rotations: 0,
            crop: None,
            color: None,
            alpha: 1.0,
            blur: 0.0,
            grayscale: false,
            invert: false,
        };

        let mut previous_filter_names: Vec<String> = Vec::new();

        for filter in filters {
            let split_filter = filter.split_once('=');
            let (filter_name, optional_raw_value) = if let Some((filter_name, raw_value)) = split_filter {
                (filter_name.to_owned(), Some(raw_value.to_owned()))
            } else {
                (filter.to_owned(), None)
            };

            if previous_filter_names.contains(&filter_name) {
                return Err(ImageFilterFromStringError::FilterUsedMoreThanOnce { name: filter_name });
            } else {
                previous_filter_names.push(filter_name.clone());
            }

            let value_is_present = optional_raw_value.is_some();
            let use_bool_value = || -> Result<bool, ImageFilterFromStringError> {
                if value_is_present {
                    Err(ImageFilterFromStringError::FilterValueIgnored {
                        filter_name: filter_name.clone(),
                    })
                } else {
                    Ok(true)
                }
            };

            let use_raw_value = || -> Result<String, ImageFilterFromStringError> {
                optional_raw_value.ok_or_else(|| ImageFilterFromStringError::FilterValueMissing {
                    filter_name: filter_name.clone(),
                })
            };

            match filter_name.as_str() {
                "crop_original" => result.crop_original = Some(parse_rect_filter_value(&filter_name, use_raw_value()?)?),
                "scale" => result.scale = parse_f32_filter_value(&filter_name, use_raw_value()?, "0..=100", Box::new(|v| (0.0..=100.0).contains(v)))?,
                "rotate" => {
                    let raw_value = use_raw_value()?;
                    let value = i32::from_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                        filter_name: filter_name.to_string(),
                        raw_value,
                    })?;

                    result.clockwise_quarter_rotations = match value {
                        0 => 0,
                        90 => 1,
                        180 => 2,
                        270 | -90 => 3,
                        _ => return Err(ImageFilterFromStringError::RotationNotAllowed),
                    };
                }
                "crop" => result.crop = Some(parse_rect_filter_value(&filter_name, use_raw_value()?)?),
                "color" => {
                    let raw_value = use_raw_value()?;

                    result.color = Some(
                        parse_rgb8_from_hex_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                            filter_name: filter_name.to_string(),
                            raw_value,
                        })?,
                    )
                }
                "alpha" => result.alpha = parse_f32_filter_value(&filter_name, use_raw_value()?, "0..<1", Box::new(|v| (0.0..1.0).contains(v)))?,
                "blur" => result.blur = parse_f32_filter_value(&filter_name, use_raw_value()?, "0..=10", Box::new(|v| (0.0..=10.0).contains(v)))?,
                "grayscale" => result.grayscale = use_bool_value()?,
                "invert" => result.invert = use_bool_value()?,
                _ => return Err(ImageFilterFromStringError::UnknownFilter { name: filter_name }),
            };
        }

        Ok(result)
    }
}
