use std::str::FromStr;

use rgb::{RGB8, RGBA8};
use serde_with::DeserializeFromStr;
use thiserror::Error;

#[derive(Debug, Error)]
#[error("The input value does not match the required format.")]
pub struct RGBParsingError {}

pub fn parse_rgb8_from_hex_str(s: &str) -> Result<RGB8, RGBParsingError> {
    let first_index = if s.starts_with('#') { 1 } else { 0 };
    if s.len() - first_index == 6 {
        let r = u8::from_str_radix(&s[first_index..(first_index + 2)], 16).map_err(|_| RGBParsingError {})?;
        let g = u8::from_str_radix(&s[(first_index + 2)..(first_index + 4)], 16).map_err(|_| RGBParsingError {})?;
        let b = u8::from_str_radix(&s[(first_index + 4)..(first_index + 6)], 16).map_err(|_| RGBParsingError {})?;

        return Ok(RGB8::new(r, g, b));
    }

    Err(RGBParsingError {})
}

pub fn parse_rgba8_from_hex_str(s: &str) -> Result<RGBA8, RGBParsingError> {
    let first_index = if s.starts_with('#') { 1 } else { 0 };
    if s.len() - first_index == 8 {
        let r = u8::from_str_radix(&s[first_index..(first_index + 2)], 16).map_err(|_| RGBParsingError {})?;
        let g = u8::from_str_radix(&s[(first_index + 2)..(first_index + 4)], 16).map_err(|_| RGBParsingError {})?;
        let b = u8::from_str_radix(&s[(first_index + 4)..(first_index + 6)], 16).map_err(|_| RGBParsingError {})?;
        let a = u8::from_str_radix(&s[(first_index + 6)..(first_index + 8)], 16).map_err(|_| RGBParsingError {})?;

        return Ok(RGBA8::new(r, g, b, a));
    }

    Err(RGBParsingError {})
}

pub fn parse_rgb8_with_optional_alpha_from_hex_str(s: &str, fallback_alpha: u8) -> Result<RGBA8, RGBParsingError> {
    // optionally +1 for the '#'
    match s.len() {
        6 | 7 => Ok(parse_rgb8_from_hex_str(s)?.alpha(fallback_alpha)),
        8 | 9 => Ok(parse_rgba8_from_hex_str(s)?),
        _ => Err(RGBParsingError {}),
    }
}

#[derive(Debug, DeserializeFromStr)]
pub struct DeserializableRGB8(pub RGB8);

impl FromStr for DeserializableRGB8 {
    type Err = RGBParsingError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        parse_rgb8_from_hex_str(s).map(|v| DeserializableRGB8(v))
    }
}

#[derive(Debug, DeserializeFromStr)]
pub struct DeserializableRGBA8(pub RGBA8);

impl FromStr for DeserializableRGBA8 {
    type Err = RGBParsingError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        parse_rgba8_from_hex_str(s).map(|v| DeserializableRGBA8(v))
    }
}

#[derive(Debug, DeserializeFromStr)]
pub struct DeserializableRGB8WithOptionalAlpha(pub RGBA8);

impl FromStr for DeserializableRGB8WithOptionalAlpha {
    type Err = RGBParsingError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        parse_rgb8_with_optional_alpha_from_hex_str(s, 0xff).map(|v| DeserializableRGB8WithOptionalAlpha(v))
    }
}
