use std::collections::HashMap;
use std::rc::Rc;

use color_eyre::eyre::{ContextCompat, WrapErr};
use color_eyre::Result;

use loupedeck_serial::commands::VibrationPattern;
use loupedeck_serial::device::LoupedeckDevice;
use state::State;

use crate::model;

mod command;
mod state;

pub fn start(config: model::config::Config) -> Result<()> {
    let state = create_state(&config)?;

    let device = LoupedeckDevice::discover()?
        .first()
        .wrap_err("No device connected.")?
        .connect()
        .wrap_err("Connecting to the device failed.")?;

    device.vibrate(VibrationPattern::RiseFall);

    Ok(())
}

fn create_state(config: &model::config::Config) -> Result<State> {
    let key_pages_by_id: HashMap<_, _> = config
        .key_pages_by_id
        .iter()
        .map(|(id, p)| state::KeyPage { id: id.clone() })
        .map(|p| (p.id.clone(), Rc::new(p)))
        .collect();

    let knob_pages_by_id: HashMap<_, _> = config
        .knob_pages_by_id
        .iter()
        .map(|(id, p)| state::KnobPage {
            id: id.clone(),
            knobs_by_position: p
                .knobs
                .iter()
                .map(|(p, k)| state::Knob {
                    id: p,
                    label: k.label.clone(),
                    icon: k.icon.clone(),
                    value: 0.0,
                })
                .map(|k| (k.id, Some(k)))
                .collect(),
        })
        .map(|p| (p.id.clone(), Rc::new(p)))
        .collect();

    Ok(State {
        active_key_page: Rc::clone(
            key_pages_by_id
                .get(&config.initial.key_page)
                .wrap_err_with(|| format!("There is no key page with the ID specified at initial.key_page: {}", &config.initial.key_page))?,
        ),
        active_knob_page: Rc::clone(
            knob_pages_by_id
                .get(&config.initial.knob_page)
                .wrap_err_with(|| format!("There is no key page with the ID specified at initial.knob_page: {}", &config.initial.key_page))?,
        ),
        key_pages_by_id,
        knob_pages_by_id,
    })
}
